from . import files, misc, text, types, phonemizer

__doc__ = """For now this will be a placeholder example:
```python
import gr1336_toolbox as gts

gts.files.find_dirs(...)
```
"""

__all__ = [
    "files",
    "types",
    "misc",
    "text",
    "phonemizer",
]
